# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
import cv2
import json
from numpy import array

from ..thresholds.color_.color import Color
from ..thresholds.color_.multi_color import MultiColor

LinearDiscriminantAnalysis = None


class HSVCalibration(object):
    """
    The calibration object that executes the machine learning
    Order: (img_mean, lower_bound)
    """
    def __init__(self, vision, class_amount, filter_min_ratio=0.9, min_area=100):
        """
        HSVCalibration object is an object that preforms a HSVCalibration function and runs during setup.
        :param vision: The Vision process to be calibration, a Vision object or Vision Json
        :param class_amount: The amount of ranges to be taken from every range for every image per class
        :param filter_min_ratio: a list with the ratio for each filter or just a float, for custom value per filter
                                 function.
        :param min_area: the minimum area a contour must have
        :returns: The calibration object
        """
        self.vision = vision
        self.color = vision.color
        self.vision.filters.remove()

        if type(filter_min_ratio) is float:
            self.rating_ratio = filter_min_ratio ** len(self.vision.get_filters())
        elif type(filter_min_ratio) is list:
            for ratio in filter_min_ratio:
                self.rating_ratio *= ratio
        self.min_area = min_area
        self.class_amount = class_amount
        self.s_clf = None
        self.v_clf = None

    class ContourRating:
        def __init__(self, contour, rating):
            self.rating = rating
            self.area = cv2.contourArea(contour)

    def brute_force(self, img, color, start_v=25, start_s=13, s_buffer=20, v_buffer=20):
        """
        Action: Calculates the Rating of each contour that passes all of the filters
        :param img: The image from which the contours should be taken
        :param color: the color that
        :param start_v: the starting v value from which to iterate, should not be 0 or negative
        :param start_s: the starting s value from which to iterate, should not be 0 or negative
        :param s_buffer: a buffer on the lower bound of color from which to check all possibilities
        :param v_buffer: a buffer on the lower bound of color from which to check all possibilities
        :return:
        """
        s_values = color.low_s + v_buffer
        v_values = color.low_v + s_buffer
        if s_values > 255 or s_values < 0:
            raise ValueError('Invalid range value,"s" must be between 0 and 255 (s = lower s + given buffer)')
        if v_values > 255 or v_values < 0:
            raise ValueError('Invalid range value,"v" must be between 0 and 255 (v = lower v + given buffer)')
        h = color.low_h
        high_range = color.high_bound
        final_list = []
        for s in range(start_s, s_values + 1):
            for v in range(start_v, v_values + 1):
                current_range = Color([h, s, v], high_range)
                self.vision.contours = self.vision.get_contours(img, color=current_range)
                self.vision.apply_all_filters(apply_all=True)
                filter_ratio_product = self.vision.apply_all_filters(apply_all=False)
                ratings = [1] * len(self.vision.contours)
                for filter_ratio in filter_ratio_product:
                    for idx, r in enumerate(filter_ratio):
                        ratings[idx] *= HSVCalibration.ContourRating(self.vision.contours[idx], r)
                ratings.sort(key=lambda k: k.rating)
                if len(ratings) >= self.vision.target_amount:
                    fail = False
                    for contour_rating in ratings[0: self.vision.target_amount]:
                        if type(contour_rating) is HSVCalibration.ContourRating:
                            if contour_rating.area <= self.min_area:
                                fail = True
                            elif contour_rating.rating < self.rating_ratio:
                                fail = True
                    if fail:
                        final_list.append([current_range.low_bound, False])
                    else:
                        final_list.append([current_range.low_bound, ratings])
                else:
                    final_list.append([current_range.low_bound, False])

        return final_list

    def find_data(self, img_list, color=None, start_v=25, start_s=13, s_buffer=20, v_buffer=20):
        """
        Action: Finds Working and non working color ranges and turns them into a list
        :param img_list: list of images from which to calibrate
        :param color: the color object
        :param start_v: the minimum valid v value
        :param start_s: the minimum valid s value
        :param s_buffer: the additional value values to check below low value
        :param v_buffer: the additional saturation values to check below low value
        :return: the Classifiers for Saturation and Value and the numpy array of data
        """
        global LinearDiscriminantAnalysis
        if LinearDiscriminantAnalysis is None:
            from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

        if color is None:
            color = self.color

        def get_data_from_color(color_obj, img_obj, img_mean):
            s_data = []
            v_data = []
            s_classifier = []
            v_classifier = []
            rating_list = self.brute_force(img_obj, color_obj, start_v, start_s, s_buffer, v_buffer)
            class1 = 0
            class2 = 0

            for obj in rating_list[0:self.class_amount]:
                if type(obj) is list:
                    bound, rating = obj
                    s_data.append([img_mean[1], bound[1]])
                    s_classifier.append(1)
                    class1 += 1
                    v_data.append([img_mean[2], bound[2]])
                    v_classifier.append(1)
                    class1 += 1

            for bound, rating in rating_list:
                if rating is False and class2 < self.class_amount:
                    s_data.append([img_mean[1], bound[1]])
                    s_classifier.append(2)
                    class2 += 1
                    v_data.append([img_mean[2], bound[2]])
                    v_classifier.append(2)
                    class2 += 1
            return s_data, s_classifier, v_data, v_classifier

        if type(color) is Color:
            saturation_data, saturation_classes, value_data, value_classes = [], [], [], []
            for img in img_list:
                tmp = get_data_from_color(color, img, cv2.mean(img))
                tmp_saturation_data = tmp[0]
                tmp_saturation_classes = tmp[1]
                tmp_value_data = tmp[2]
                tmp_value_class = tmp[3]
                saturation_data.extend(tmp_saturation_data)
                saturation_classes.extend(tmp_saturation_classes)
                value_classes.extend(tmp_value_class)
                value_data.extend(tmp_value_data)
            if not (len(saturation_data) == len(saturation_classes) == len(value_classes) == len(value_data)):
                raise ValueError("Data isn't equal in size!")
            if saturation_data is []:
                raise ValueError("No data")
            self.s_clf = LinearDiscriminantAnalysis(solver='eigen')
            self.s_clf.fit(array(saturation_data), array(saturation_classes))
            self.v_clf = LinearDiscriminantAnalysis(solver='eigen')
            self.v_clf.fit(array(value_data), array(value_classes))

        elif type(color) is MultiColor:
            saturation_data, saturation_classes, value_data, value_classes = [], [], [], []
            for img in img_list:
                mean = cv2.mean(img)
                for color in color.colors:
                    tmp = get_data_from_color(color, img, mean)
                    tmp_saturation_data = tmp[0]
                    tmp_saturation_classes = tmp[1]
                    tmp_value_data = tmp[2]
                    tmp_value_class = tmp[3]
                    saturation_data.extend(tmp_saturation_data)
                    saturation_classes.extend(tmp_saturation_classes)
                    value_classes.extend(tmp_value_class)
                    value_data.extend(tmp_value_data)
            if not (len(saturation_data) == len(saturation_classes) == len(value_classes) == len(value_data)):
                raise ValueError("Data isn't equal in size!")
            if saturation_data is []:
                raise ValueError("No data")
            saturation_data, saturation_classes = array(saturation_data), array(saturation_classes)
            value_data, value_classes = array(value_data), array(value_classes)
            self.s_clf = LinearDiscriminantAnalysis(solver='eigen')
            self.s_clf.fit(saturation_data, saturation_classes)
            self.v_clf = LinearDiscriminantAnalysis(solver='eigen')
            self.v_clf.fit(value_data, value_classes)
            return self.s_clf, self.v_clf

    def json_serialize(self, file_name=None):
        """
        Action: Turns a Trained HSVCalibration object to a Json File for use in a Vision program
                And saves it as a json file
        :return: The Translation of the Object to Python
        """
        s_wc1, s_wc2 = self.s_clf.coef_[0]
        s_w_intercept = self.s_clf.intercept_
        s_mean = self.s_clf.means_

        v_wc1, v_wc2 = self.v_clf.coef_[0]
        v_w_intercept = self.v_clf.intercept_
        v_mean = self.v_clf.means_

        translation = {
            'saturation_weight_vector': {'coefficient1': s_wc1,
                                         'coefficient2': s_wc2,
                                         'inter': s_w_intercept,
                                         'mean': s_mean},
            'brightness_weight_vector': {'coefficient1': v_wc1,
                                         'coefficient2': v_wc2,
                                         'inter': v_w_intercept,
                                         'mean': v_mean}}
        if file_name is not None:
            if not file_name.endswith('.json'):
                print(None, 'Target file is not a json file')
            with open(file_name, 'r') as f:
                f.write(json.dumps(translation))
        return translation



import numpy as np
import cv2


def validate_odd_size(size):
    """
     validates that a kernel shape is of odd  ints and of size 2
    :param size: the shape (size) to be checked
    :return: doesnt raise an error if it's ok.
    """
    if type(size) not in (list, tuple):
        return False
    if len(size) != 2:
        return False
    if size[0] % 2 != 1 or size[1] % 2 != 1:
        return False
    return True


def is_odd_size(size) -> bool:
    """
    Validates that a kernel shape is of odd  ints and of size 2
    :param size: the shape (size) to be checked
    :return: doesnt raise an error if it's ok.
    """
    if type(size) not in (list, tuple):
        return False
    if len(size) != 2:
        return False
    if size[0] % 2 != 1 or size[1] % 2 != 1:
        return False
    return True


def cross_kernel(size):
    """
    Returns a cross (ones in a cross) kernel for morphological functions
    Example of a (5,5) cross:
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    | 1 1 1 1 1 |
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    :param size:  a tuple of size 2 of 2 odd integers denoting the size of the kernel
    f.g. (5, 5)
    :return: the numpy.array of the cross shape
    """
    validate_odd_size(size)
    return cv2.getStructuringElement(cv2.MORPH_CROSS, ksize=size)


def rectangle_kernel(size):
    """
    Returns a rectangle (all ones) kernel for morphological functions
    Example of a (5,5) rectangle:
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    :param size:  a tuple of size 2 of 2 odd integers denoting the size of the kernel
    f.g. (5, 5)
    :return: the numpy.array of the cross shape
    """
    return cv2.getStructuringElement(cv2.MORPH_RECT, ksize=size)


def ellipse_kernel(size):
    """
    Returns an ellipse (ones in the shape of an ellipse) kernel for morphological functions
    Example of a (5,5) ellipse:
    | 0 0 1 0 0 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 1 1 1 1 1 |
    | 0 0 1 0 0 |
    :param size: a tuple of size 2 of 2 odd integers denoting the size of the kernel
    f.g. (5, 5)
    :return: the kernel
    """
    validate_odd_size(size)
    return cv2.getStructuringElement(cv2.MORPH_ELLIPSE, ksize=size)


def horizontal_line_kernel(size):
    """
    Returns an horizontal line (a horizontal line of ones) kernel for morphological functions
    Example of a (5,5) horizontal line:
    | 0 0 0 0 0 |
    | 0 0 0 0 0 |
    | 1 1 1 1 1 |
    | 0 0 0 0 0 |
    | 0 0 0 0 0 |
    :param size: a tuple of size 2 of 2 odd integers denoting the size of the kernel
    f.g. (5, 5)
    :return: the kernel
    """
    validate_odd_size(size)
    kernel = np.zeros(size, dtype=np.uint8)
    kernel[int((size[0] - 1) / 2), ] = 1
    return kernel


def vertical_line_kernel(size):
    """
    Returns a vertical line (a vertical line of ones) kernel for morphological functions
    Example of a (5,5) vertical line:
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |
    | 0 0 1 0 0 |

    :param size: a tuple of size 2 of 2 odd integers denoting the size of the kernel
    f.g. (5, 5)
    :return: the kernel
    """
    validate_odd_size(size)
    kernel = np.zeros(size, dtype=np.uint8)
    kernel[:, int((size[1] - 1) / 2)] = 1
    return kernel



import typing

VisionLike = typing.Union["Vision", "AmbientVision"]


def RangedNumber(start, end):
    class Hint:
        def __class_getitem__(cls, item):
            cls.start = item[0]
            cls.end = item[1]
            return cls
    return typing.Union[float, Hint[start, end]]



from functools import reduce
import time
import math
import copy
import types
import cv2
import numpy as np
import json
from typing import List, Union, Tuple, Any

from ..thresholds.color_ import built_in_colors
from ..exceptions_.exceptions import InvalidCustomFunctionError, CameraError
from ..camera_.camera import Camera
from ..partials.filter_applier import filter_applier
from ..thresholds.threshold import Threshold
from ..thresholds.color_.color import Color
from ..connections.connection import Connection
from ..directions_.director import Director
from ..camera_.camera_settings import CameraSettings
from ..connections.network_location import NetworkLocation
from ..directions_.directing_functions import center_directions


class Vision:
    """
    The Vision object represents the specifications to detect an object in an image.
    It performs the detection using a Threshold and filter functions.

    Additional capabilities and tuning options are:
        Image filters (Blurs, rotations, cropping),
        Morphological functions,
        Ovl color HSVCalibration,


    """
    def __init__(self, threshold: Threshold = None, contour_filters: List[types.FunctionType] = None,
                 director: Director = None, width=320, height=240, connection: Connection = None,
                 camera: Union[int, str, Camera, cv2.VideoCapture, Any] = None,
                 camera_settings: CameraSettings = None, morphological_functions: List[types.FunctionType] = None,
                 image_filters: List[types.FunctionType] = None, ovl_camera: bool = False, calibration: str = None):
        """
        The object that represents the pipeline of processing, detection, direction of values
        A connection object can be passed
        :param threshold: threshold object, creates the binary mask from a given image
        :param contour_filters: the list of contour_filter functions that
                                remove contours that aren't the target(s)
        :param director: a functions that receives a list or a single contour and returns director
        :param width: the width (in pixels) of images taken with the camera
        :param height: the height (in pixels)
        :param connection: a connection object that passes the result to the connection target
        :param camera: a Camera object (cv2.VideoCapture, ovl.Camera) or source from which to open a camera
        :param camera_settings: Special camera settings like calibration or offset used for
                                image correction and various direction calculations.
        :param morphological_functions: morphological functions used to apply on the binary
                               mask generated by the Threshold.
        :param image_filters: a list of image altering functions that are applied on the image.
        :param ovl_camera: a boolean that makes the camera opened to be ovl.Camera instead of cv2.VideoCapture
        :param calibration: a dictionary containing color calibration (HSVCalibration) coefficients and intercepts,
                            used for applying the calibration on the vision object
        """
        self.width = width
        self.height = height
        self.threshold = threshold
        self.contour_filters = contour_filters or []
        self.director = director or Director(center_directions, failed_detection=9999, target_amount=1)
        self.connection = connection
        self.image_filters = image_filters or []
        self.morphological_functions = morphological_functions or []
        self.camera = None
        self.camera_port = None
        self.camera_settings = camera_settings
        if isinstance(camera, (cv2.VideoCapture, Camera)):
            self.camera = camera
        elif camera is None:
            pass
        else:
            self.camera_setup(camera, width, height, ovl_camera=ovl_camera)

        self.calibration_path = None
        if calibration:
            self.calibration_path = calibration
            calibration = json.load(calibration)
            self.saturation_weight = calibration['saturation'] if 'saturation' in calibration else None
            self.brightness_weight = calibration['brightness'] if 'brightness' in calibration else None

    def __repr__(self):
        return str(self)

    def __str__(self):
        filters = [filter_function.__name__ for filter_function in self.contour_filters]
        threshold = self.threshold
        return "Vision: \n Threshold: {} \n Filters: {}".format(threshold, filters)

    @property
    def target_amount(self):
        """
        The wanted amount of targets
        (0 None or math.inf if there is no limit, 1 if 1 target is wanted etc.)
        """
        if self.director is None:
            return math.inf
        return self.director.target_amount

    def apply_morphological_functions(self, mask, morphological_functions=None):
        """
        Applies all morphological functions on the mask
        :param mask: the mask on which the functions should be applied
        :param morphological_functions: list of morphological_functions to be
                             applied instead of self.morphological_functions
        :return: the applied mask
        """
        if type(self.morphological_functions) not in (tuple, list, set):
            return mask
        morphological_functions = morphological_functions or self.morphological_functions
        return reduce(filter_applier, morphological_functions, mask)

    def send(self, data, *args, **kwargs) -> Union[None, Any]:
        """
        Sends data to the destination.
        :param data: The data to send to the Connection
        :param args: any other arguments for the send function in your connection
        :param kwargs: any other named arguments for the connection object
        :return: True if data was successfully sent, False if not.
        """
        return self.connection.send(*args, **kwargs, data=data)

    def send_to_location(self, data, network_location: NetworkLocation, *args, **kwargs):
        return self.connection.send_to_location(data, network_location, *args, **kwargs)

    def get_image(self) -> np.ndarray:
        """
        Gets an image from self.camera and applies image filters
        :return: the image, false if failed to get it
        """
        if self.camera is None:
            raise CameraError("No camera given, (Camera is None)")
        if not self.camera.isOpened():
            raise CameraError("Camera given is not open (Has it been closed or disconnected?)")
        output = self.camera.read()
        if len(output) == 2:
            ret, image = output
            return image if ret else False
        else:
            return output

    def get_filtered_image(self):
        """
        Gets an image from self.camera and applies all image filters
        :return: the image filter applied image
        """
        output = self.get_image()
        return self.apply_image_filters(output)

    def apply_filter(self, filter_function, contours, verbose=False):
        """
        Applies a filter function of the contour list
        :param filter_function: Filter functions are function with a contour list variable that apply some
        sort of filter on the contours, thus removing ones that don't fit the limit given by the filter.
        for example: straight_rectangle_filter removes contours that are not rectangles that are parallel to the frame
        of the picture
        :param contours: the contours on which the filter should be applied (list of numpy.ndarrays)
        :param verbose: if true_shape does not print anything
        :return: returns the output of the filter function.
        """
        if verbose:
            print('Before "{}": {}'.format(filter_function.__name__, len(contours)))
        filter_function_output = filter_function(contours)

        if isinstance(filter_function_output, tuple):
            if len(filter_function_output) == 2:
                filtered_contours, ratio = filter_function_output[0], filter_function_output[1]
            else:
                raise InvalidCustomFunctionError('Filter function must return between 1 and 2 lists.'
                                                 'Please refer to the Documentation: '
                                                 'https://github.com/1937Elysium/Ovl-Python')
        elif isinstance(filter_function_output, list):
            filtered_contours, ratio = filter_function_output, []
        else:
            raise TypeError('The contour list must be a list or tuple of 2 lists (contours and ratios)')
        return filtered_contours, ratio

    def apply_all_filters(self, contours: List[np.ndarray], verbose=False
                          ) -> Tuple[List[np.ndarray], List[float]]:
        """
         Applies all of the filters on a list of contours
        :param contours: List of contours (numpy arrays) to
        :param verbose: prints out information about filtering process if true (useful for debugging)
        :return: a list of all of the ratios given by the filter function in order.
        """
        ratios = []
        for filter_func in self.contour_filters:
            contours, ratio = self.apply_filter(filter_func, contours, verbose=verbose)
            ratios.append(ratio)
        if verbose:
            print("After all filters: {}".format(len(contours)))
        return contours, ratios

    def apply_image_filters(self, image: np.ndarray) -> np.ndarray:
        """
         applies all given image filters to the given image
        :param image: the image that the image filters should be applied on (numpy array)
        :return: the image with the filters applied
        """
        return reduce(filter_applier, self.image_filters, image)

    def apply_threshold(self, image: np.ndarray, threshold=None) -> np.ndarray:
        """
         gets a mask for a given img and Threshold object (uses self.Threshold if given threshold was none)
        :param image: the numpy array of the image
        :param threshold: the Threshold object used to create the binary mask
        :return: the binary mask
        """
        threshold = threshold or self.threshold
        return threshold.convert(image)

    def find_contours_in_mask(self, mask: np.ndarray, return_hierarchy=False, apply_morphs=True
                              ) -> List[np.ndarray]:
        """
         gets contours from the given mask and apply
        :param mask: binary image (mask), a numpy array
        :param return_hierarchy: if the hierarchy should be returned
        :param apply_morphs: if the morphological functions should be applied.
        :return: the list of contours
        """
        mask = self.apply_morphological_functions(mask) if apply_morphs else mask
        result = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        if len(result) == 3:
            _, contours, hierarchy = result
        elif len(result) == 2:
            contours, hierarchy = result
        else:
            raise ValueError("Invalid output from cv2.findContours, check that your cv2 (OpenCV) version is supported")
        return (contours, hierarchy) if return_hierarchy else contours

    def find_contours(self, image: np.ndarray, threshold=None, return_hierarchy=False) -> List[np.ndarray]:
        """
         Gets a list of all the contours within the threshold that was given
        :param threshold: the threshold that determines the binary masks
        :param image: image from which to get the contours
        :param return_hierarchy: if the hierarchy should be returned
        :return: list of all contours matching the range of hsv colours
        :rtype: list
        """
        threshold = threshold or self.threshold
        image_mask = self.apply_threshold(image, threshold=threshold)
        return self.find_contours_in_mask(image_mask, return_hierarchy=return_hierarchy)

    def get_directions(self, contours: List[np.ndarray], image: np.ndarray, sorter=None):
        """
        Calculates the directions, based on contours found in the given image
        :param contours: final contours after filtering
        :param image: the image from which to find the contours
        :param sorter: optional parameter, applies a sorter on the given contours
        :return: a string of the director (output of the director function),
                 length depends on the director function
        """
        return self.director.direct(contours, image, camera_settings=self.camera_settings, sorter=sorter)

    def camera_setup(self, source=0, image_width=None, image_height=None, ovl_camera=False):
        """
         Opens up the camera reference and sets a given width and height to all images taken
        :param image_width: the width of the images to be taken, 0 does not set a width
        :param image_height: the height of the images to be taken, 0 does not set a height
        :param source: the location from which to open the camera
                       string for network connections int for local USB connections.
        :param ovl_camera: if the camera object should be ovl.Camera
        :return: the camera object, also sets self.camera to the object.
        """
        image_height = image_height or self.height
        image_width = image_width or self.width
        self.camera_port = source
        if ovl_camera:
            robot_cam = Camera(source=source, image_width=image_width, image_height=image_height)
        else:
            robot_cam = cv2.VideoCapture(source)
            if image_width != -1:
                robot_cam.set(3, image_width)
            if image_height != -1:
                robot_cam.set(4, image_height)

        if not robot_cam.isOpened():
            raise CameraError("Camera did not open correctly! Camera source: {}".format(self.camera_port))
        self.camera = robot_cam
        return robot_cam

    def apply_on_sample(self, image: Union[np.ndarray, str], threshold: Threshold = None, display_result: bool = False,
                        delay=0, result_color: Color = built_in_colors.RED_HIGH_HSV, return_hierarchy=False
                        ) -> Tuple[List[np.ndarray], np.ndarray]:
        """
         Finds contours and applies filters on a single image given by a image path or image object (numpy array)
        :param image: an image on which to apply the vision object
        :param threshold: a custom threshold can be passed
        :param display_result: if the contours found should be displayed
        :param delay: waitKey delay (0 if not in a loop, non-zero if it is)
        :param result_color: the colour of the contour outline
        :param return_hierarchy: if the contour hierarchies should be return
        :return: the image (numpy array) and the contours found (list of numpy arrays)
                 Can also return the hierarchies if return_hierarchy is true
        """
        threshold = threshold or self.threshold
        image = cv2.imread(image) if type(image) == str else image
        contours, hierarchy = self.find_contours(image, threshold, return_hierarchy)
        if display_result:
            image_for_display = copy.copy(image)
            cv2.drawContours(image_for_display, contours, -1, result_color)
            cv2.imshow(str(time.time()), image_for_display)
            cv2.waitKey(delay)
        return (contours, image, hierarchy) if return_hierarchy else (contours, image)

    def detect(self, image, verbose=False, return_ratios=False):
        """
        Gets contours and applies all filters and returns the result, thus detecting the object according
        to the specifications in the vision
        :param image: image in which the vision should detect an object
        :param verbose: If information about the filtering should be printed
        :param return_ratios: if the ratios from the filters should be returned
        :return: contours and the filtered image and the ratios if return_ratios is true
        """
        image = self.apply_image_filters(image)
        contours = self.find_contours(image)
        contours, ratios = self.apply_all_filters(contours, verbose)
        return (contours, image, ratios) if return_ratios else (contours, image)



def validate_loaded_vision(loaded_vision):
    """
     Verifies the a loaded vision (Dictionary) contains the the necessary
            Attributes to create a Vision object
    :param loaded_vision: A dictionary containing raw data of the Vision object
    :return:
    """
    return loaded_vision


def loads_vision(serialized_vision):
    """
    Loads a Vision object from a json string.
    :param serialized_vision:
    :return:
    """
    raise NotImplementedError()
    # loaded_vision = loads(serialized_vision)
    # if not validate_loaded_vision(loaded_vision):
    #     raise ValueError("The vision loaded was incomplete")


def load_vision(vision_file):
    """
        Loads a Vision object from a json string.
        :param vision_file: path to the vision file (a json file)
        :return:
        """
    raise NotImplementedError()
    # with open(vision_file, "r") as vision:
    #     return load_serialized_vision(loads(vision.read()))



def calculate_calibrated_value(image_mean, vector):
    """
     Solves the calibration equation that finds the optimal low bound value for
            the saturation and value.
    :param image_mean: the mean if the image of which
    :param vector: the dictionary containing the coefficients and group mean.
     Calculated using Color HSVCalibration
    :return: the optimal low bound
    """
    data_mean = vector['mean'][0]
    z_mean = data_mean[0] * vector['coefficient1'] + data_mean[1] * vector['coefficient2']
    return (z_mean - (image_mean * vector['coefficient1'])) / vector['coefficient2']



# Copyright 2018 - 2019 Ori Ben-Moshe - All rights reserved.
from functools import reduce

import cv2
import numpy as np
import typing
import math
from typing import List, Tuple, Union

from .default_math_functions import math_funcs


def target_size(contours: typing.List[np.ndarray]) -> float:
    """
    Returns the sum of contour areas
    """
    return sum(map(cv2.contourArea, contours))


def y_intersection(line_slope, intercept, x_val):
    """
     finds the y value of the line (y = mx + b) at point x and returns the point
    :param line_slope: slope of the line (m)
    :param intercept: the intercept of the line (b)
    :param x_val: the value to be used (substituted with x)
    :return: the y value
    """
    return x_val, line_slope * x_val + intercept


def x_intersection(line_slope, intercept, y_val):
    """
     calculates the x value of which the line according to the given y value
    :param line_slope:
    :param intercept:
    :param y_val:
    :return:
    """
    return (y_val - intercept) / float(line_slope), y_val if line_slope != 0 else (0, y_val)


def distance_from_frame(point, image_dimensions):
    """
    Action calculates the distance of a given point from the frame of the image based on the vector from the center
    and the point
    :param point: point (x,y tuple) or contour (numpy array)
    :param image_dimensions: the size of the image, (width, height)
    :return: the distance
    """
    point = contour_center(point) if type(point) is np.ndarray else point
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    line_slope = slope(point, image_center)
    intercept = - line_slope * point[0] + point[1]
    xs = (image_dimensions[0] - 1, 0)
    ys = (image_dimensions[1] - 1, 0)
    y_distances = [distance_between_points(y_intersection(line_slope, intercept, y), point) for y in ys]
    x_distances = [distance_between_points(x_intersection(slope, intercept, x), point) for x in xs]
    return min(y_distances + x_distances)


def horizon_angle(point: Union[np.ndarray, typing.Tuple[int, int]], field_of_view: float, image_width: int) -> float:
    """
     returns the angle compared to the center of the image for a given field of view image width and point
    :param point: the contour or its center
    :param field_of_view: the horizontal field of view
    :param image_width: the image width
    :return: the angle compared to the center of the image , negative left positive right.
    """
    point = point[0] if type(point) in (tuple, list, set) else point
    x_value = contour_center(point)[0] if type(point) is np.ndarray else point
    angle = math.atan((x_value - (image_width - 1) / 2) / float(focal_length(image_width, field_of_view)))
    return float(math.degrees(angle))


def vertical_angle(point: Union[np.ndarray, typing.Tuple[int, int]], field_of_view: float, image_height: int) -> float:
    """
     returns the angle compared to the center of the image for a given field of view image height and point
    :param point: the x val or center
    :param field_of_view: the vertical field of view
    :param image_height: the image height
    :return: the angle compared to the center of the image , negative up positive down6.
    """
    point = point[1] if type(point) in (tuple, list, set) else point
    y_val = contour_center(point)[1] if type(point) is np.ndarray else point
    angle = math.atan((y_val - (image_height - 1) / 2) / float(focal_length(image_height, field_of_view)))
    return float(math.degrees(angle))


def focal_length(image_width, field_of_view):
    """
    Calculates the focal length in pixels of the camera for a given an image width and field of view
    :param image_width: width of the image in pixels
    :param field_of_view:
    :return: calculates the focal length, a float
    """
    return image_width / float((2 * math.tan(math.radians(float(field_of_view / 2)))))


def open_arc_length(contour: np.ndarray) -> float:
    """
     Returns the arc length of an "open" contour
    :param contour:
    :return:
    """
    return cv2.arcLength(contour, False)


def slope(first_point, second_point):
    """
     returns the slope between 2 points
    :param first_point: first point (x,y)
    :param second_point: second point (x,y)
    :return: the slope
    """
    return 0. if first_point[0] == second_point[0] else\
        ((float(first_point[1]) - float(second_point[1])) / (float(first_point[0]) - float(second_point[0])))


def angle_between_points(first_point, second_point):
    """
     calculates the angle between the 2 points and the horizontal axis
    :param first_point: the first point (2d)
    :param second_point: the second point  (2d)
    :return: the angle
    """
    return math.degrees(math.atan(slope(first_point, second_point)))


def line_angle(contour):
    """
     returns the angle of the line created by the top-most and
    bottom-most points of a contour
    :param contour: the contour, a numpy array
    :return: the angle in degrees
    """
    if type(contour) != np.ndarray:
        raise TypeError("Contour must be a numpy array")
    topmost = tuple(contour[contour[:, :, 1].argmin()][0])
    bottommost = tuple(contour[contour[:, :, 1].argmax()][0])
    return angle_between_points(topmost, bottommost)


def open_contour_approximation(contour: np.ndarray, approximation_coefficient: float = 0.02):
    """
     returns the vertices of the approximation of the contour
    NOTE: for "open" (unclosed shape) contours only!
    :param contour: open contour (numpy array)
    :param approximation_coefficient: approximation coefficient
    :return: list of vertices
    """
    arc = approximation_coefficient * cv2.arcLength(contour, False)
    return cv2.approxPolyDP(contour, arc, False)


def calculate_math_expression(equation, value, additional_functions=None, symbol='x'):
    """
     calculates the result of a math expression with a variable
    :param equation: the equation with a variable to be calculated
    :param value: the value to be placed instead of the variable
    :param additional_functions: additional functions to be used
    :param symbol: The symbol in the equation representing the variable
    :return: the result of the variable
    """
    expression = equation.replace(symbol, str(value))
    if additional_functions is None:
        additional_functions = math_funcs

    exec('equation_res = ' + expression, additional_functions)
    return locals()['equation_res']


def circle_area(radius):
    """
     calculates the area of a circle with the given radius.
    :param radius: The radius of the circle
    :return: Area of the circle
    """
    return math.pi * (radius ** 2)


def distance_between_points(first_point, second_point):
    """
     Calculates the Distance between 2 points. (x^2 + y^2) ^ 0.5
    :param first_point: tuple of x and y value of point 1
    :param second_point: tuple of x and y value of point 2
    :return: Float value of the distance between the 2 points
    :rtype: float
    """
    return ((first_point[0] - second_point[0]) ** 2 + (first_point[1] - second_point[1]) ** 2) ** 0.5


def rectangle_fill_ratio_straight(contour, reverse_div=False):
    """
     Returns the ratio between the contour and the straight bounding rectangle.
    :param contour: The contour to be compared
    :param reverse_div: If the rectangle's area should be divided by the contour.
    :return:  the ratio, the width of the bounding rectangle, the height of the bounding rectangle
    :rtype: float
    """
    _, _, width, height = cv2.boundingRect(contour)
    bounding_rectanlge_area = width * height
    contour_area = cv2.contourArea(contour)
    if reverse_div:
        return float(bounding_rectanlge_area) / contour_area, width, height
    return float(contour_area) / bounding_rectanlge_area, width, height


def triangle_fill_ratio(contour, triangle=None) -> float:
    """
    :param contour: numpy array
    :param triangle: convex hull for the minEnclosingTriangle function, can be ignored
    :return: returns the ratio between the contour and the bounding triangle
    :rtype: float
    """
    bounding_triangle_area, triangle = cv2.minEnclosingTriangle(contour, triangle)
    contour_area = cv2.contourArea(contour)
    return contour_area / float(bounding_triangle_area)


def rotating_rectangle_fill_ratio(contour: np.ndarray) -> typing.Tuple[float, float, float]:
    """
     Returns the ratio between the contour and the smallest bounding rectangle.
    :param contour: The contour to be compared
    :return: the ratio, the width of the bounding rectangle, the height of the bounding rectangle
    :rtype: float, float, float
    """
    rotated = cv2.minAreaRect(contour)
    bounding_rectangle_area = rotated[1][0] * rotated[1][1]
    contour_area = cv2.contourArea(contour)
    return float(contour_area) / bounding_rectangle_area, rotated[1][0], rotated[1][1]


def circle_fill_ratio(contour: np.ndarray) -> typing.Tuple[float, float]:
    """
     Returns the ratio between the contour and the smallest enclosing circle.
    :param contour: The contour to be compared
    :return: the ratio, the radius of the enclosing circle
    :rtype: float
    """
    _, enclosing_radius = cv2.minEnclosingCircle(contour)
    enclosing_circle_area = circle_area(enclosing_radius)
    contour_area = cv2.contourArea(contour)
    return float(contour_area) / enclosing_circle_area, enclosing_radius


def contour_center(contour: np.ndarray) -> Tuple[float, float]:
    """
     Returns the coordinates center of a contour
    :param contour: a single contour
    :return: X coordinate of center, Y coordinate of center
    """
    moments = cv2.moments(contour)
    try:
        area = float(moments['m00'])
        center_x, center_y = moments['m10'] / area, moments['m01'] / area
    except ZeroDivisionError:
        raise ValueError('Contour given is too small!,'
                         'try using area_filter to remove small contours (try min_area=20)')
    return center_x, center_y


def _contour_center_sum(point_sum, contour):
    """
    Helper function for contour_average_center
    """
    def _average_point_reduce(first_point, second_point):
        return first_point[0] + second_point[0], first_point[1] + second_point[1]
    current_contour_center = contour_center(contour)
    return _average_point_reduce(point_sum, current_contour_center)


def contour_average_center(contours) -> Tuple[float, float]:
    """
    Calculates the average center of a list of contours
    :param contours: the list of contours
    :return: the average center  (x,y)
    """
    return reduce(_contour_center_sum, contours, (0, 0))


def regular_polygon_angle(side_amount: int) -> float:
    """
     returns the inner angle of a polygon with the given amount of sides
    :param side_amount: amount of sides the polygon has
    :return:
    """
    return (side_amount - 2) * (180 / side_amount)


def contour_approximation(contour, approximation_coefficient=0.02):
    """
     returns the approximation of the contour
    :param contour: a numpy array
    :param approximation_coefficient: the coefficient of the contour length approximation
    :return:
    """
    perimeter = cv2.arcLength(contour, True)
    approximation = cv2.approxPolyDP(contour, approximation_coefficient * perimeter, True)
    return approximation


def polygon_area(side_length, amount_of_sides):
    """
     area of a polygon of
    :param side_length: side_length of a side of the polygon
    :param amount_of_sides: Amount of sides the polygon has
    :return: the area of the polygon
    """
    angle = 180 / amount_of_sides
    return 0.25 * amount_of_sides * (side_length ** 2) * (math.cos(math.radians(angle)) / math.sin(math.radians(angle)))


def circle_rating(contour, area_factor=0.9, radius_factor=0.8):
    """
     returns a rating of how close is the circle to being a circle
    :param contour: the contour that its rating is calculated
    :param area_factor: the factor (p-value) that separates an area ratio that is a circle and one that isn't
    :param radius_factor: the factor (p-value) that separates an radius ratio that is a circle and one that isn't
    :return:
    """
    fill_ratio, radius = circle_fill_ratio(contour)
    _, _, width, height = cv2.boundingRect(contour)
    radius_ratio = ((((radius * 2) ** 2) / float(width) * height) ** 0.5)
    rating = (radius_ratio * radius_factor) * (fill_ratio * area_factor)
    return rating


def law_of_cosine(first_point, second_point, third_point) -> float:
    """
    Finds the angle BAC  in degrees given that first_point = A second point = B and third point = C
    """
    first_length = distance_between_points(first_point, second_point)
    second_length = distance_between_points(first_point, third_point)
    third_length = distance_between_points(second_point, third_point)
    angle = (first_length ** 2 + second_length ** 2 - third_length ** 2) / (first_length * second_length * 2)
    return math.degrees(angle)


def contour_lengths_and_angles(contour, approximation_coefficient=0.02):
    """
    Returns a list of lengths and angles of a given contour.
    :param contour: contour (numpy array) to find its
    :param approximation_coefficient:
    :return: returns the list of sides
    """
    approximation = contour_approximation(contour, approximation_coefficient)
    vertices = []
    lengths = []
    angles = []
    for points in approximation:
        vertex = points[0][1], points[0][0]
        vertices.append(vertex)

    for index, point in enumerate(vertices):
        if index != len(approximation) - 1:
            current_length = distance_between_points(point, vertices[index + 1])
            current_angle = law_of_cosine(vertices[index - 1], point, vertices[index + 1])
        else:
            current_length = distance_between_points(point, vertices[0])
            current_angle = law_of_cosine(vertices[index - 1], point, vertices[0])
        lengths.append(current_length)
        angles.append(current_angle)

    return vertices, lengths, angles



import cv2
import numpy as np

from .image_filter import image_filter
from ..helpers_.remove_none_values import remove_none_values
from .kernels import validate_odd_size


def convert_to_hsv(img):
    """
     converts an image to hsv - Mainly for beginner use
    :param img: image to be converted
    :return: the converted image
    """
    return cv2.cvtColor(img, cv2.COLOR_BGR2HSV)


@image_filter
def sharpen_image(image, size=(3, 3)):
    """
     Sharpens an image by preforming convolution it with a sharpening matrix
    :param image: the image (numpy array)
    :param size: the size of the sharpening matrix
    :return: the new sharpened image
    """

    validate_odd_size(size)
    kernel = np.ones(size)
    kernel *= -1
    kernel[int((size[0] - 1) / 2), int((size[1] - 1) / 2)] = kernel.size
    return cv2.filter2D(image, -1, kernel)


@image_filter
def adaptive_brightness(image, brightness=127, hsv=False):
    """
     Changes the brightness of every pixel so that the average is the target average
    :param image: The image to be changed (Numpy array)
    :param brightness: the target average for the image
    :type brightness: int (int between 0 - 100)
    :param hsv: bool noting if the image is in hsv
    :return: a copy of the image changed
    :rtype: numpy array
    """
    image = image if hsv else cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    brightness = brightness * 2.55
    image_hue_mean = cv2.mean(image)[2]
    increase = brightness - image_hue_mean
    vid = image[:, :, 2]
    if increase > 0:
        vid = np.where(vid + increase <= 255, vid + increase, 255)
    else:
        vid = np.where(vid + increase >= 0, vid + increase, 0)
    image[:, :, 2] = vid
    image = image if hsv else cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
    return image


@image_filter
def change_brightness(image, change=25, hsv_image=False):
    """
     Changes the brightness of every pixel of a BGR image by the given amount
    :param image: The image to be changed (Numpy array)
    :param change: the change (integer) (The min brightness is 0 and max is 100)
    :type change: int
    :param hsv_image: bool noting if the image is in hsv
    :return: a copy of the image changed
    """
    image = image if hsv_image else cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    change = change * 2.55
    vid = image[:, :, 2]
    if change > 0:
        vid = np.where(vid + change <= 255, vid + change, 255)
    else:
        vid = np.where(vid + change >= 0, vid + change, 0)
    image[:, :, 2] = vid
    image = image if hsv_image else cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
    return image


@image_filter
def rotate_image(image, angle=180):
    """
    Rotates an image by a given amount of degrees.
    Note that the rotated image's dimensions will most likely stay the same if the angle is not 90, -90 or 180
    :param image: the image to be rotated
    :param angle: the angle to rotate in (positive is to the left, negative to the right)
    :return: the rotated image
    """
    angle = angle % 360
    shortcut_angles = {
        90: rotate90_left,
        -90: rotate90_right,
        180: rotate180,
        -180: rotate180
    }
    if angle in shortcut_angles:
        return shortcut_angles[angle]()(image)
    elif angle == 0:
        return image
    else:
        return rotate_by_angle()(image, angle)


@image_filter
def rotate_by_angle(image, angle):
    """
    Rotates the given image by a given angle
    :param image: the image to be rotated
    :param angle: the angle of rotation
    :return: the rotated image
    """
    (height, width) = image.shape[:2]
    center_xy = (width / 2, height / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center_xy, -angle, 1.0)
    cos = np.abs(rotation_matrix[0, 0])
    sin = np.abs(rotation_matrix[0, 1])
    rotated_image_dimensions = (int((height * sin) + (width * cos)), int((height * cos) + (width * sin)))
    rotation_matrix[0, 2] += (rotated_image_dimensions[0] / 2) - center_xy[0]
    rotation_matrix[1, 2] += (rotated_image_dimensions[1] / 2) - center_xy[1]
    return cv2.warpAffine(image, rotation_matrix, rotated_image_dimensions)


@image_filter
def rotate90_left(image):
    """
    Return a copy of the image rotated 90 degrees to the left (_counter-clockwise)
    :param image: numpy array, image to be rotated
    :return: a copy of the image rotated.
    """
    (height, width) = image.shape[:2]
    center = (width / 2, height / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center, 90, 1.0)
    return cv2.warpAffine(image, rotation_matrix, (height, width))


@image_filter
def rotate90_right(image):
    """
    Return a copy of the image rotated 90 degrees to the right (clockwise)
    :param image: numpy array, image to be rotated
    :return:
    """
    (height, width) = image.shape[:2]
    center = (width / 2, height / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center, 270, 1.0)
    return cv2.warpAffine(image, rotation_matrix, (height, width))


def rotate180(image):
    """
    Return a copy of the image rotated 180 degrees
    :param image:  numpy array, image to be rotated
    :return: a copy of the image rotated.
    """
    (height, width) = image.shape[:2]
    center = (width / 2, height / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center, 180, 1.0)
    return cv2.warpAffine(image, rotation_matrix, (height, width))


@image_filter
def non_local_mean_denoising(image, h, h_color, template_window_size, search_window_size, destination=None):
    parameters = {"h": h,
                  "hColor": h_color,
                  "templateWindowSize": template_window_size,
                  "seasonWindowSize": search_window_size,
                  "dst": destination
                  }
    return cv2.fastNlMeansDenoisingColored(image,
                                           **remove_none_values(parameters))


@image_filter
def gaussian_blur(image, kernel_size=(3, 3), sigma_x=5, sigma_y=None, border_type=None):
    parameters = {"ksize": kernel_size,
                  "sigmaX": sigma_x,
                  "sigmaY": sigma_y,
                  "borderType": border_type}
    return cv2.GaussianBlur(image, **remove_none_values(parameters))


@image_filter
def crop_image(image, point, dimensions):
    """
    Crops a given rectangle from a given image
    :param image: an image (numpy array)
    :param point: (x, y) coordinates that is the top left corner of the rectangle
    :param dimensions: (width, height) of the rectangle
    :return: the region of the image
    """
    x, y = point
    width, height = dimensions
    return image[x: x + width, y: y + height]



import cv2
from numpy import vstack, hstack, ndarray

WINDOWS = []


def stitch_images(images):
    width, height = images[0].shape
    stitched_image = cv2.imread(images[0]) if type(images[0]) is str else images[0]
    for image in images[1:]:
        if type(image) is str:
            image = cv2.imread(image)
        image_shape = image.shape
        if width + image_shape[0] > height + image_shape[1]:
            stitched_image = vstack((stitched_image, image))
            width += image_shape[0]
        else:
            stitched_image = hstack((stitched_image, image))
            height += image_shape[1]
    return stitched_image


def show_image(image, window_name, delay):
    cv2.imshow(window_name, image)
    pressed_key = cv2.waitKey(delay)
    return pressed_key


def display_image(image, window_name='image', delay=0, resizable=False):
    """
     The function displays the image whose path is received.
    :param image: Represents an image path (string), an already open image in the for of a numpy array (ndarray)
                  or a list of images (strings and arrays are valid)
    :param window_name: Name of the Window that displays the images
    :param delay: The delay before the image stops
    :param resizable: A boolean that determines whether the image window is resizable or not
    :return: the key pressed while image was displayed
    """
    if type(image) is str:
        image = cv2.imread(image)
    elif type(image) is list:
        image = stitch_images(image)
    elif type(image) is not ndarray:
        raise TypeError("Invalid image given, image must be an image, a path to an image or a list of images")

    if resizable:
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)

    return show_image(image, window_name, delay)





import time


def numeral_name_convention(image_number):
    return "image" + str(image_number)


def time_name_convention(image_number):
    return "image{}:{}".format(image_number, time.strftime('%a-%d-%b-%Y-%H:%M:%S', time.gmtime()))



from typing import List
import numpy as np
import functools

from ..partials.reverse_partial import ReversePartial


def _loaded_condition(loaded_contour_filter, contour_list: List[np.ndarray]):
    return list(filter(loaded_contour_filter, contour_list))


def conditional_contour_filter(single_contour_filter):
    """
    A contour_filter that turns a function that filters a single contour (and returns true if it has passed)
    to an iterative one that goes over a list

    example:
        @conditional_contour_filter
        def area_filter(contour, minimum_area=50)
            return cv2.contourArea(contour) > minimum_area

    This function can then be passed to a Vision object:
        filters = [area_filter(min_area=60)] # sets the minimum area to 60 -> the first parameter passed
        vision = Vision(..., contour_filters=filters, ...)

    :param single_contour_filter: the function to turn into a contour filter, which loads the parameters to the filter
    before running it on list of contour filters.
    :return: the argument loader that wraps the target function
    """

    def argument_loader(*args, **kwargs):
        condition = ReversePartial(single_contour_filter, *args, **kwargs)
        argument_loader.condition = condition
        return functools.partial(_loaded_condition, condition)
    return argument_loader



import cv2

from ..contour_filter import contour_filter
from ...math_ import geometry
from ...helpers_.types import RangedNumber


@contour_filter
def polygon_filter(contour_list, side_amount=6, min_angle_ratio: RangedNumber(0, 1) = 0.7,
                   min_area_ratio: RangedNumber(0, 1) = 0.7, min_len_ratio: RangedNumber(0, 1) = 0.7,
                   approximation_coefficient: RangedNumber(0, 1) = 0.02):
    """
     A filter that Detects regular polygon of n-sides sides
    :param contour_list: the list of contours to be filtered
    :param side_amount: the amount of sides the wanted polygon has.
    :param min_angle_ratio: the minimum ratio between the each angle and the average angle and the average angle and the
                            target angle of a shape of side_amount
    :param min_area_ratio: The minimum ratio between the contour's area and the target area
    :param min_len_ratio: The minimum ratio between the length of each side and the average length and
                          between the average length.
    :param approximation_coefficient: the coefficient for the function cv2.approxPolyDP.
    :return: the list of contours (numpy ndarrays) that passed the filter and are regular polygon
    """
    if side_amount < 3:
        raise ValueError("A polygon must have at least 3 sides")
    output = []
    ratios = []
    output_append = output.append
    ratio_append = ratios.append
    for current_contour in contour_list:
        vertices, lengths, angles = geometry.contour_lengths_and_angles(current_contour, approximation_coefficient)
        if len(vertices) == side_amount:
            target_angle = geometry.regular_polygon_angle(side_amount)
            average_length = round(sum(lengths) / float(len(lengths)), 3)
            average_angle = round(sum(angles) / float(len(lengths)), 3)
            for length in lengths:
                if 1 - ((((length - average_length) ** 2) ** 0.5) / average_length) < min_len_ratio:
                    break
            else:
                for angle in angles:
                    if 1 - ((((angle - average_angle) ** 2) ** 0.5) / average_angle) < min_angle_ratio:
                        break
                else:
                    if not 1 - ((((target_angle - average_angle) ** 2) ** 0.5) / target_angle) < min_angle_ratio:
                        ratios = cv2.contourArea(current_contour) / geometry.polygon_area(average_length, side_amount)
                        if min_area_ratio <= ratios <= 1 / min_area_ratio:
                            ratio_append(ratios)
                            output_append(current_contour)
    return output, ratios



from .shape_filters import *
from .conditional_contour_filter import conditional_contour_filter
from .contour_filter import contour_filter, CONTOUR_FILTERS
from .sorters import *
from .contour_filters import *



from ..partials.keyword_partial import keyword_partial


def image_filter(image_filter_function):
    """
    A decorator used to pass parameters in two stages,
    run in a python console "help(ovl.contour_filter)" or ovl.contour_filter.__doc__
    or look at ovl.contour_filter documentation for more information
    image_filter acts like contour_filter except it receives an image (np.ndarray)
    applies changes on it and then returns it altered.
    functions decorated by image_filter can be passed to Vision(image_filters_)

    Changes that can be applied vary from various blurs like Gaussian blur or Bilateral filter to image rotations
    translations, contrast and brightness alterations and color spaces changes (HSV, RGB Greyscale conversions)

    Filters that change binary images (images that have been threshold by
    a threshold object such as ovl.Color ovl.MultiColor or ovl.Canny) like morphological
    functions (ex. erosion, dilation) count as image filters as well and should be
    decorated using image_filter.

    image filter example:
        1 |    @image_filter
        2 |    def rotate_by_angle(image, angle):
        3 |        (height, width) = image.shape[:2]
        4 |        center_xy = (width / 2, height / 2)
        5 |        rotation_matrix = cv2.getRotationMatrix2D(center_xy, -angle, 1.0)
        6 |        cos = np.abs(rotation_matrix[0, 0])
        7 |        sin = np.abs(rotation_matrix[0, 1])
        8 |        rotated_image_dimensions = (int((height * sin) + (width * cos)), int((height * cos) + (width * sin)))
        9 |        rotation_matrix[0, 2] += (rotated_image_dimensions[0] / 2) - center_xy[0]
        10|        rotation_matrix[1, 2] += (rotated_image_dimensions[1] / 2) - center_xy[1]
        11|        return cv2.warpAffine(image, rotation_matrix, rotated_image_dimensions)
    """
    return keyword_partial(image_filter_function)



import cv2

from ..contour_filter import contour_filter
from ...math_.geometry import rectangle_fill_ratio_straight
from ...helpers_.types import RangedNumber


@contour_filter
def straight_square_filter(contour_list, min_area_ratio: RangedNumber(0, 1) = 0.8,
                           min_len_ratio: RangedNumber(0, 1) = 0.95):
    """
    Receives a list of contours and returns
    only the ones that are approximately square
    Relation checked is [minimum ratio < (Circle radius / width * height * (square root of 2)) < maximum ratio]
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_len_ratio: maximum ratio between radius and sides of bounding rectangle
    :param min_area_ratio: the minimum ratio between the area of the contour and the area of the bounding shape
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    for current_contour in contour_list:
        fill_ratio, contour_width, contour_height = rectangle_fill_ratio_straight(current_contour)
        peri = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and len(approximation) == 4:
            _, enclosing_radius = cv2.minEnclosingCircle(current_contour)
            bounding_rectangle_diagonal = (2 ** 0.5) * ((contour_width * contour_height) ** 2)
            enclosing_circle_radius = 2 * enclosing_radius
            radius_ratio = enclosing_circle_radius / bounding_rectangle_diagonal
            if radius_ratio >= min_len_ratio:
                output_list.append(current_contour)
                ratio_list.append(fill_ratio * radius_ratio)
    return output_list, ratio_list



import typing
import numpy as np

from ...math_ import geometry
from ..direction_monitors.direction_monitor import DirectionMonitor


class StopIfCloseMonitor(DirectionMonitor):
    """
    A Direction monitor that stops when the target found is large enough (close enough).
    Like other DirectionMonitor objects, it monitors the directions returned.
    """
    def __init__(self, minimum_size: float, value_sent: typing.Any, priority: bool = False):
        """
        :param minimum_size: This is the size in pixels that considers the object "close" enough
                            it depends on your object size, but 40% of the image size is usually enough
                            so for an image the size of 320 x 240 it will be 320 * 240 * 0.4
        :param value_sent: the value returned if the object is close enough
        :param priority: a boolean that notes if this monitor should take priority (and stop following monitors from
                         being called)
        WARNING: Setting priority to true can cause 'unexpected' behaviour as a result of stopping
        """
        self.minimum_size = minimum_size
        self.value_sent = value_sent
        self._priority = priority

    @property
    def priority(self):
        return self._priority

    def monitor(self, directions: typing.Any, target_contours: typing.List[np.ndarray],
                image: np.ndarray, mask: np.ndarray) -> typing.Any:
        if geometry.target_size(target_contours) >= self.minimum_size:
            return self.value_sent
        return directions



from ..partials import keyword_partial


CONTOUR_FILTERS = set()


def contour_filter(contour_filter_function):
    """
    A decorator function used to make a contour filter function.
    Contours filters are functions that take a list of contours
    and returns only


    Wrapping a function it gives the following capabilities:
        - pass parameters before the list of contours
        - preserves the properties of the function: name, documentation etc.

    The contours filter then changes it behaviour to act as follows:


    For a given function:

        @contour_filter
        def area_filter(contours, min_area, max_area):
            output_list = []
            ratio_list = []

            for current_contour in contour_list:
                if min_area <= cv2.contourArea(current_contour) <= max_area:
                    output_list.append(current_contour)
                    ratio_list.append(current_contour)
            return output_list, ratio_list


    Instead of calling the function like other functions:

        A1| some_filter(list_of_contours, 200, 5000)

    The function needs to be called as follows:

        B1| activator = area_filter(200, 5000)
        B2|
        B3| result = activator(list_of_contours)


    Vision objects use functions that are decorated with contour_filter
    you can just pass the result of line B1 to the Vision object
    like so:

        C1| contour_filters = [area(400, 5000), ovl.circle_filter(0.75)]
        C2|
        C3| vision = Vision(..., contour_filters=contour_filters, ...)
    """
    CONTOUR_FILTERS.add(contour_filter_function)
    return keyword_partial.keyword_partial(contour_filter_function)



import cv2

from ..contour_filter import contour_filter
from ...math_.geometry import triangle_fill_ratio
from ...helpers_.types import RangedNumber


@contour_filter
def triangle_filter(contour_list, min_area_ratio: RangedNumber(0, 1) = 0.8,
                    approximation_coefficient: RangedNumber(0, 1) = 0.02):
    """
    Receives a list of contours and returns only those that are approximately
    triangle and have 3
    :param approximation_coefficient: the approximation coefficient affects
    how the contour's perimeter is approximated, the larger the number is (out of 1) the more the approximation of
    the contour is a line
    :param contour_list: the list of contours to be filtered
    :param min_area_ratio: the minimum ratio between the area of the contour and the bounding triangle
    :return: the filtered list
    """
    output_list = []
    ratio_list = []
    for current_contour in contour_list:
        fill_ratio = triangle_fill_ratio(current_contour)
        peri = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, approximation_coefficient * peri, True)
        if fill_ratio > min_area_ratio and len(approximation) == 3:
            ratio_list.append(fill_ratio)
            output_list.append(current_contour)
    return output_list, ratio_list



from functools import partial

from ..math_ import geometry


def contour_and_point_distance(point):
    """
    A reusable function that preloads a point and then
    calculates the distance to a given contour.
    :param point: an (x,y) tuple point
    :return:
    """
    return partial(geometry.distance_between_points, point)



from ..math_.geometry import *
from ..math_.geometry import distance_between_points as dbp, distance_from_frame as dff
from .contour_filter import contour_filter
from ..helpers_.types import RangedNumber


@contour_filter
def image_center_filter(contour_list, image_dimensions: typing.Tuple[int, int] = (320, 240),
                        max_dist: RangedNumber(0, 1) = 0.7):
    """
    Filters out contours that their center is not close enough to the center of the image
    :param contour_list: a list of contours to be filtered
    :param image_dimensions: the size of the image(width, height)
    :param max_dist: the maximum percent of difference
    :return: list of contours within the maximum distance from the image center
    """
    output = []
    ratio = []
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    for current_contour in contour_list:
        current_contour_center = contour_center(current_contour)
        distance_ratio = (dbp(current_contour_center, image_center)
                          / float(dff(current_contour_center, image_dimensions)))
        if distance_ratio >= max_dist:
            output.append(current_contour)
            ratio.append(distance_ratio)
    return output, ratio


@contour_filter
def distance_filter(contour_list, point: typing.Tuple[int, int], min_dist: float = 0, max_dist: float = 50):
    """
    Filters out contours that their center is not close enough
    to the given (x, y) point in the image
    :param contour_list: a list of contours to be filtered
    :param point: the point from which the contours should be filtered a tuple (or list) of 2 numbers.
    :param max_dist: the maximum distance from the point in pixels
    :param min_dist: the minimum distance from the point in pixels
    :return: the filtered contour list
    """
    output = []
    ratio = []
    for current_contour in contour_list:
        current_contour_center = contour_center(current_contour)
        distance_from_center = dbp(current_contour_center, point)
        if max_dist >= distance_from_center >= min_dist:
            output.append(current_contour)
            ratio.append(distance_from_center)
    return output, ratio


@contour_filter
def absolute_distance_filter(contour_list, max_dist=50, min_dist=0, image_dimensions=(320, 240)):
    """
    Filters out contours that their center is not close enough
    to the center of the image
    :param contour_list: a list of contours to be filtered
    :param image_dimensions: the size of the image(width, height)
    :param max_dist: the maximum distance from the center in pixels
    :param min_dist: the minimum distance from the center in pixels
    :return:
    """
    output = []
    ratio = []
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    for current_contour in contour_list:
        current_contour_center = contour_center(current_contour)
        distance_from_center = dbp(current_contour_center, image_center)
        if max_dist >= distance_from_center >= min_dist:
            output.append(current_contour)
            ratio.append(distance_from_center)
    return output, ratio


@contour_filter
def length_filter(contour_list, min_length=50, max_length=76800):
    """
    Receives a list of contours and removes ones that are not long enough
    Note: for "open" contours only!
    :param contour_list: list of contours (numpy array) to be filtered
    :param min_length: minimum length of a contour (in pixels)
    :param max_length: maximum length of a contour (in pixels)
    :return: list of filtered contours and list of the lengths
    """
    output = []
    ratio = []
    for current_contour in contour_list:
        perimeter = cv2.arcLength(current_contour, False)
        if min_length >= perimeter >= max_length:
            output.append(current_contour)
            ratio.append(perimeter)
    return output, ratio


@contour_filter
def area_filter(contour_list, min_area=200, max_area=math.inf):
    """
    Filters contours that are not within the threshold of area (in pixels)
    :param max_area: maximum area of a contour (Inclusive) default is no limit (infinity)
    :param min_area: minimum area of a contour (Inclusive) set to 0 for no lower limit.
    :param contour_list: List of Contours to filter
    :return: the contour list filtered.
    """
    output_list = []
    ratio_list = []
    for current_contour in contour_list:
        if min_area <= cv2.contourArea(current_contour) <= max_area:
            output_list.append(current_contour)
            ratio_list.append(current_contour)
    return output_list, ratio_list


@contour_filter
def percent_area_filter(contour_list, minimal_percent: RangedNumber(0, 1) = 0.002,
                        maximum_percent: RangedNumber(0, 1) = 1, image_dimensions: Tuple[int, int] = (320, 240)):
    """
    Filters out contours that are not in the specified ratio to the area of the image (1% -> 0.1)
    :param contour_list: list of contours to be filtered (numpy.ndarray)
    :param minimal_percent: the minimal ratio between the contour area and the image area
    :param maximum_percent: the maximum ratio between the contour area and the image area
    :param image_dimensions: The (width, height) of the image in pixels,
    :return:
    """
    output = []
    ratios = []
    output_append = output.append
    ratio_append = ratios.append
    image_size = image_dimensions[0] * image_dimensions[1]
    if image_size == 0:
        raise ValueError("Invalid image dimensions, Received (width, height): {}, {}".format(*image_dimensions))
    for current_contour in contour_list:
        contour_area = cv2.contourArea(current_contour)
        percent_area = contour_area / image_size
        if minimal_percent <= percent_area <= maximum_percent:
            ratio_append(percent_area)
            output_append(current_contour)
    return output, ratios



import numpy as np
from typing import Union, List, Any, Generator, Tuple, Dict

from ..helpers_.types import VisionLike
from .ambient_vision import AmbientVision
from ..connections.connection import *
from ..connections.network_location import NetworkLocation


class MultiVision:
    """
    An object used to switch between multiple vision objects.
    It does this by having a list of visions and a connection and network_location to update from
    So it can be automatically updated to swap to the desired Vision object
    use the start method to start an infinite loop that returns images and detections

        for contours, image, directions in multi_vision.start():
            # do something with the generated data
            # like sending the data or displaying the contours
            multi_vision.send(directions)
            ovl.display_contours(image)
    """

    def __init__(self, visions: Union[List[VisionLike], Dict[Any, VisionLike]], update_connection: Connection,
                 update_location: Union[NetworkLocation, None] = None, default_vision=0):
        self.current = visions[default_vision]
        self.index = default_vision
        self.visions = visions
        self.connection = update_connection
        self.update_location = update_location
        self.switch_functions = {list: self._list_switch_vision,
                                 dict: self._dict_switch_vision}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()

    def is_ambient(self) -> bool:
        """
         Returns True if the current_vision vision is an Ambient vision object
        :return:
        """
        return isinstance(self.current, AmbientVision)

    def send(self, data, *args, **kwargs) -> Any:
        """
        Sends the value to the target destination of the current_vision vision object
        :param data: the data to be sent
        :return:
        """
        return self.current.send(data=data, *args, **kwargs)

    def set_vision(self, index):
        """
        Sets the current to the given index
        :param index: the index to set
        :return: the index
        """
        self.current = self.visions[self.index]
        self.index = index
        return index

    def _dict_switch_vision(self, index: int):
        """
        Switches the current vision to the one given if the visions container is a list
        """
        return index in self.visions

    def _list_switch_vision(self, index: Any):
        """
        Switches the current vision to the one given if the visions container is a dictionary
        """
        return 0 <= index < len(self.visions)

    def switch_vision(self, index: Any):
        """
        Switches the
        :param index: the index of the new vision, can be an int if the container of the visions
        is a list or any immutable object if it is a dictionary
        :return: the index set (the index given if it is valid and
        """
        if self.switch_functions[type(self.visions)](index):
            self.set_vision(index)
            return index
        return self.index

    def update_current(self) -> int:
        """
        Reads the updated current vision from the update network location
        and then updates the current vision
        :return: the index received
        """
        new_idx = self.connection.receive_from_location(network_location=self.update_location or {})
        self.switch_vision(new_idx)
        return new_idx

    def start(self, yield_ratios=False) -> Generator[Tuple[List[np.ndarray], np.ndarray, Any], None, None]:
        """
        A function that starts an infinite generator that takes an image
        detects with the current vision and returns the list of contours the image and directions
        and should be used as follows:
             for directions, contours, image in multi_vision.start():
                 # do something with the generated data
                 # like sending the data or displaying the contours
                 multi_vision.send(directions)
                 ovl.display_contours(image)
        :param yield_ratios: if True also yields the list of ratios returned from
        :yields: contours image directions and ratios if yield_ratios if True
        """
        while True:
            self.update_current()
            filtered_image = self.current.get_filtered_image()
            detection_result = self.current.detect(filtered_image, return_ratios=yield_ratios)
            contours = detection_result[0]
            image = detection_result[1]
            directions = self.current.director.direct(contours, image, self.current.camera_settings)
            if self.is_ambient():
                self.current.update_vision()
            yield (directions, *detection_result)



import functools


class ReversePartial(functools.partial):
    """
     A keyword_partial that passes the call arguments before the loaded (passed in constructor) arguments
    """
    def __call__(*args, **keywords):
        if not args:
            raise TypeError("descriptor '__call__' of keyword_partial needs an argument")
        self, *args = args
        new_keyword_arguments = self.keywords.copy()
        new_keyword_arguments.update(keywords)
        return self.func(*args, *self.args, **new_keyword_arguments)



from .circle_filter import circle_filter
from .constraining_circle_filter import constraining_circle_filter
from .horizontal_rectangle_filter import horizontal_rectangle_filter
from .polygon_filter import polygon_filter
from .rotated_square_filter import rotated_square_filter
from .straight_square_filter import straight_square_filter
from .vertical_rectangle_filter import vertical_rectangle_filter
from .rotated_rectangle_filter import rotated_rectangle_filter
from .straight_rectangle_filter import straight_rectangle_filter
from .triangle_filter import triangle_filter



from socket import inet_aton, error


def is_static_ip(ip):
    """
    Checks if the given ip is a static ip for FRC
    :param ip: the ip to be checked, a str
    :return: True if the ip is an FRC static ip, False if it isn't
    """
    if not isinstance(ip, str):
        return False
    if len(ip) == 10:
        if ip.startswith('10.') and ip.count('.') == 3:
            return True
    return False


def is_valid_ip(ip):
    """
     checks if the given ip address (string) is a valid ip
    :param ip: the string of the ip to be checked, can be ipv6 and ipv4
    :return: False if invalid, True if valid
    """
    try:
        inet_aton(ip)
        return True
    except error:
        return False



import cv2

from ..contour_filter import contour_filter
from ...math_.geometry import rectangle_fill_ratio_straight
from ...helpers_.types import RangedNumber


@contour_filter
def horizontal_rectangle_filter(contour_list, min_area_ratio: RangedNumber(0, 1) = 0.8):
    """
    Receives a list of contours and returns only those that are approximately a horizontal rectangle
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_area_ratio
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    for current_contour in contour_list:
        fill_ratio, contour_width, contour_height = rectangle_fill_ratio_straight(current_contour)
        perimeter = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * perimeter, True)
        if fill_ratio > min_area_ratio and contour_width > contour_height and len(approximation) == 4:
            output_list.append(current_contour)
            ratio_list.append(fill_ratio)
    return output_list, ratio_list



import typing


def team_number_to_ip(team_number: typing.Union[str, int]):
    """
    Turns a team number into the static ip of the roborio
    TEAM -> 10.TE.AM.2
    :param team_number: string or int of the team number
    """

    team_number = str(team_number)
    first_numbers = team_number[:-2]
    second_numbers = team_number[-2:]
    if first_numbers == "":
        first_numbers = 0
    team_number = "{}.{}".format(first_numbers, second_numbers)
    return "10.{team_number}.2".format(team_number=team_number)



import cv2
import json
import numpy as np

from ..helpers_.remove_none_values import remove_none_values
from .threshold import Threshold


class CannyEdge(Threshold):
    """
    A Canny edge detection Threshold
    Creates a binary image using the Canny edge detection algorithm
    See:
    https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_imgproc/py_canny/py_canny.html
    """
    def __init__(self, low: int, high: int, aperture_size: int = None, l2_gradient: bool = None):
        self.low = low
        self.high = high
        self.aperture_size = aperture_size
        self.l2_gradient = l2_gradient

    def validate(self) -> bool:
        """
        Validate that the Canny edge parameters are valid
        """
        return 255 > self.high > self.low > 0

    def canny_parameters(self):
        """
        Returns the parameters of the Canny threshold that are not None
        """
        parameters = {"apertureSize": self.aperture_size,
                      "L2gradient": self.l2_gradient}
        return remove_none_values(parameters)

    def convert(self, image: np.ndarray) -> np.ndarray:
        return cv2.Canny(image, self.low, self.high, **self.canny_parameters())

    def serialize(self):
        """
        Serializes the Canny Threshold to a json string
        """
        return json.dumps({"low": self.low,
                           "high": self.high,
                           "aperture_size": self.aperture_size,
                           "l2_gradient": self.l2_gradient})

    @staticmethod
    def deserialize_canny_threshold(serialized_canny_threshold: str):
        canny_dictionary = json.loads(serialized_canny_threshold)
        return CannyEdgeThreshold(**canny_dictionary)



import cv2

from ..contour_filter import contour_filter
from ...math_.geometry import rotating_rectangle_fill_ratio
from ...helpers_.types import RangedNumber


@contour_filter
def rotated_square_filter(contour_list, min_area_ratio: RangedNumber(0, 1) = 0.8, min_ratio: RangedNumber(0, 1) = 0.95,
                          max_ratio: RangedNumber(0, 1) = 1.05):
    """
    Receives a list of contours
    and returns only the ones that are approximately square
    Relation checked is [minimum ratio < (Circle radius / width * height * (square root of 2)) < maximum ratio]
    :param contour_list: List of Contours to filter
    :param max_ratio: maximum ratio between radius and sides of bounding rectangle
    :type max_ratio: float
    :param min_ratio: minimum ratio between radius and sides of bounding rectangle
    :type min_ratio: float
    :param min_area_ratio: minimum ratio between the area of the contours and the bounding shape
    :return: the contour list filtered.
     """
    ratio_list = []
    output_list = []
    for current_contour in contour_list:
        fill_ratio, bounding_width, bounding_height = rotating_rectangle_fill_ratio(current_contour)
        _, enclosing_radius = cv2.minEnclosingCircle(current_contour)
        contour_perimeter = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * contour_perimeter, True)
        if fill_ratio > min_area_ratio and len(approximation) == 4:
            diagonal_length = (2 ** 0.5) * ((bounding_width * bounding_height) ** 0.5)
            enclosing_circle_diameter = 2 * enclosing_radius
            radius_ratio = enclosing_circle_diameter / diagonal_length
            if min_ratio < radius_ratio < max_ratio:
                output_list.append(current_contour)
                if radius_ratio > 1:
                    radius_ratio = 1 / radius_ratio
                ratio_list.append(fill_ratio * radius_ratio)
    return output_list, ratio_list



import cv2

from ..contour_filter import contour_filter
from ...math_.geometry import circle_fill_ratio
from ...helpers_.types import RangedNumber


@contour_filter
def constraining_circle_filter(contour_list, min_area_ratio: RangedNumber(0, 1) = 0.80,
                               min_len_ratio: RangedNumber(0, 1) = 0.9):
    """
    Filters out contour which are not approximately circle, also limits by radius ratio (unlike circle_filter).
    :param contour_list: list of contours (numpy arrays) to be filtered
    :param min_area_ratio: minimum ratio between the area of the enclosing
                      circle and the contour (contour /enclosing circle)
    :param min_len_ratio: minimum ratio between the radius of the enclosing circle and the enclosing rectangle.
    :return: the list of contours that fit the criteria
    """
    output = []
    ratios = []
    output_append = output.append
    ratio_append = ratios.append
    for current_contour in contour_list:
        fill_ratio, radius = circle_fill_ratio(current_contour)
        _, _, rectangle_width, rectangle_height = cv2.boundingRect(current_contour)
        radius_ratio = ((2 * radius) ** 2) / float(rectangle_width) * rectangle_height
        if min_len_ratio <= (radius_ratio ** 0.5):
            if min_area_ratio <= fill_ratio:
                output_append(current_contour)
                ratio_append((fill_ratio, radius_ratio))
    return output, ratios



from ..contour_filter import contour_filter
from ...math_.geometry import circle_fill_ratio
from ...helpers_.types import RangedNumber


@contour_filter
def circle_filter(contour_list, min_area_ratio: RangedNumber(0, 1) = 0.7):
    """
    Filters out contour which are not approximately circle.
    :param contour_list: list of contours (numpy arrays) to be filtered
    :param min_area_ratio: minimum ratio between the area of the enclosing
                         circle and the contour (contour /enclosing circle)
    :return: the list of contours that fit the criteria
    """
    output = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, radius = circle_fill_ratio(current_contour)
        if min_area_ratio <= fill_ratio:
            output.append(current_contour)
            ratio_list.append(fill_ratio)
    return output, ratio_list



import cv2
import warnings

from ..math_ import geometry
from .contour_filter import contour_filter


@contour_filter
def area_sort(contour_list, descending_sort=True):
    """
    Sorts the list of contours by contour sort, default is from large to small
    :param contour_list: the list of contours to be sorted, list of ndarrays
    :param descending_sort: a flag that reverses the sort order set to
    :return: the sorted contour list
    """
    return sorted(contour_list, key=lambda contour: cv2.contourArea(contour), reverse=descending_sort)


@contour_filter
def distance_sort(contour_list, point):
    """
     Sorts the contours according to their distance from the
    NOTE: it is important to area filter before image_center_sort
    :param contour_list: The list of contours to be sorted
    :type contour_list: List or one contour (numpy array)
    :param point: the point from which the distance of all contours are sorted by
    :return: the sorted contour list
    """
    if type(contour_list) is not list:
        return [contour_list]
    return sorted(contour_list,
                  key=lambda x: geometry.distance_between_points(point, geometry.contour_center(x)))


@contour_filter
def image_center_sort(contour_list, image_dimensions=(320, 240)):
    """
     sorts the contours from the closest to the center to the farthest.
    NOTE: it is important to area filter before image_center_sort
    :param contour_list: list of contours to be sorted
    :type contour_list: List or one contour (numpy array)
    :param image_dimensions: the dimensions of the image
    :return:
    """
    if type(contour_list) is not list:
        return [contour_list]
    image_center = ((image_dimensions[0] - 1) / 2, (image_dimensions[1] - 1) / 2)

    return sorted(contour_list,
                  key=lambda x: geometry.distance_between_points(image_center, geometry.contour_center(x)))


@contour_filter
def dec_area_sort(contour_list):
    """
     sorts the list of contours from the largest to the smallest based on area of the contour
    :param contour_list: List of Contours to be sorted
    :type contour_list: List or one contour (numpy array)
    :return: the contour list sorted.
    """
    warnings.warn("dec_area_sort is deprecated, use area_sort() instead", DeprecationWarning)
    if type(contour_list) is not list:
        return [contour_list]
    return sorted(contour_list, key=lambda x: cv2.contourArea(x), reverse=True)


@contour_filter
def inc_area_sort(contour_list):
    warnings.warn("inc_area_sort is deprecated, use area_sort(descending_sort=False) instead", DeprecationWarning)
    """
     sorts the list of contours from the smallest to the largest based on area of the contour
    :param contour_list: List of Contours to filter
    :type contour_list: List or one contour (numpy array)
    :return: the contour list sorted.
    """
    if type(contour_list) is not list:
        return [contour_list]
    return sorted(contour_list, key=lambda x: cv2.contourArea(x))


@contour_filter
def circle_sort(contour_list, area_limit=0.9, radius_limit=0.8):
    """
     sorts the list of contours according to how much they are circle from most similar to least
            using the circle rating function.
    :param contour_list: list of Contours to filter
    :type contour_list: list or one contour(numpy array)
    :param area_limit: The area limit for the circle rating, look at Geometry.circle_rating
    :param radius_limit: the radius limit for the circle rating, look at Geometry.circle_rating
    :return: the list sorted by circle rating
    """
    if type(contour_list) is not list:
        return [contour_list]
    return sorted(contour_list, key=lambda x: geometry.circle_rating(x, area_limit, radius_limit))


@contour_filter
def dec_length_sort(contour_list):
    """
     sorts the list of contours from the largest to the smallest based on area of the contour
    :param contour_list: List of Contours to filter
    :type contour_list: List or one contour (numpy array)
    :return: the contour list sorted.
    """
    if type(contour_list) is not list:
        return [contour_list]
    return sorted(contour_list, key=geometry.open_arc_length).reverse()


@contour_filter
def inc_length_sort(contour_list):
    """
     sorts the list of contours from the smallest to the largest based on area of the contour
    :param contour_list: List of Contours to filter
    :type contour_list: List or one contour (numpy array)
    :return: the contour list sorted.
    """
    if type(contour_list) is not list:
        return [contour_list]
    return sorted(contour_list, key=geometry.open_arc_length)



from inspect import getmembers, isfunction


def get_defined_methods():
    """
     Returns a list of all available functions within the current_vision scope
    :return: list of function object that are callable
    """
    function_list = []
    for module, module_object in globals().items():
        for sub_object_name, sub_object in getmembers(module_object):
            if isfunction(sub_object):
                function_list.append((sub_object_name, sub_object))
                continue
            for sub_members_name, sub_members_objects in getmembers(sub_object):
                if isfunction(sub_members_objects):
                    function_list.append((sub_members_name, sub_members_objects))
    return function_list
